import shutil
from enum import Enum
from pathlib import Path
from typing import Dict, Optional, Tuple

import rtoml
from dynaconf import Dynaconf  # type: ignore
from pydantic import BaseSettings, FilePath, root_validator

THIS_DIR = Path(__file__).parent.absolute()


class Scope(str, Enum):
    builtin = "builtin"
    user = "user"
    local = "local"


def to_snake_case(d: dict) -> dict:
    """Recursively converts the keys of a dictionary to snake_case style."""
    res = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = to_snake_case(v)

        res[k.lower().replace("-", "_")] = v

    return res


def to_str(d: dict) -> dict:
    """Recursively converts the values of a dictionary to str."""
    res = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = to_str(v)
        else:
            v = str(v)

        res[str(k)] = v

    return res


class Settings(BaseSettings):
    pipenv: Optional[FilePath]
    pipx: Optional[FilePath]
    poetry: Optional[FilePath]
    spack: Optional[FilePath]
    __dynaconf_settings__: Optional[Dynaconf]

    @root_validator
    def which_path(cls, values):
        paths = {"pipenv", "pipx", "poetry", "spack"}

        for p in paths:
            if values.get(p, None) is None:
                if _ := shutil.which(p):
                    values[p] = Path(_)

        return values

    @staticmethod
    def settings_scopes() -> Dict[Scope, Path]:
        return {
            Scope.builtin: THIS_DIR / "settings.toml",
            Scope.user: Path("~/.config/pyvarium/settings.toml").expanduser(),
            Scope.local: Path("./pyvarium.toml").absolute(),
        }

    @classmethod
    def load_dynaconf(cls) -> "Settings":
        """Load configurations via Dynaconf and parse them into a Settings object."""

        cls.__dynaconf_settings__ = Dynaconf(
            includes=list(cls.settings_scopes().values()),
            environments=False,
        )

        # TODO: there is a planned feature for dynaconf to allow defining schemas with
        # pydantic directly, which would avoid this weird dynaconf -> dict -> dict with
        # different keys -> pydantic steps
        __dynaconf_dict__ = to_snake_case(cls.__dynaconf_settings__.to_dict())  # type: ignore

        return cls.parse_obj(__dynaconf_dict__)

    def write(self, scope: Scope = Scope.local):
        settings = self.dict()
        target = self.settings_scopes()[scope].expanduser().absolute()
        target.parent.mkdir(parents=True, exist_ok=True)
        target.write_text(rtoml.dumps(to_str(settings)))


settings = Settings.load_dynaconf()



import os
import subprocess
import sys
from pathlib import Path
from typing import Dict, Optional, Union

from loguru import logger
from rich.status import Status

from pyvarium.config import settings


class Program:
    executable: Path
    cwd: Path
    env: Union[Dict, os._Environ]

    def __init__(
        self,
        executable: Optional[Path] = None,
        post_init: bool = True,
        status: Optional[Status] = None,
    ):
        if status:
            original_text = status.status
            update_text = f"{original_text}: {{x}}"
            self.update_status = lambda x: status.update(update_text.format(x=x))
        else:
            self.update_status = lambda *_: None

        if executable is None:
            executable = settings.__getattribute__(self.__class__.__name__.lower())

        if executable is None:
            raise ValueError(f"{self.__class__.__name__} executable not found")

        self.executable = Path(executable)

        self.cwd = Path.cwd()
        self.env = os.environ.copy()
        self.env.clear()
        # FIX: something somewhere in pipenv/python test PATH to `None`, which then
        # causes an exception in `subprocess.run`
        python_bin = str(Path(sys.executable).absolute().parent)
        self.env["PATH"] = f"{python_bin}:/usr/local/bin:/usr/bin:/bin"

        if post_init:
            self.__post_init__()

    def __post_init__(self):
        ...

    def cmd(self, *args) -> subprocess.CompletedProcess:
        logger.debug(f"`{self.executable.name} {' '.join(args)}`")
        self.update_status(f"`{self.executable.name} {' '.join(args)}`")
        res = subprocess.run(
            [self.executable, *args],
            cwd=self.cwd,
            env=self.env,
            capture_output=True,
        )

        logger.debug(res)

        if res.returncode != 0:
            raise RuntimeError(f"Process return code is not 0: {res=}")

        return res

    def config(self):
        ...

    @property
    def version(self) -> str:
        out = self.cmd("--version").stdout

        if type(out) is bytes:
            return out.decode().strip()
        else:
            return out.strip()


class Environment:
    program: Program

    def __init__(
        self,
        path: Path,
        program: Optional[Program] = None,
        post_init: bool = True,
        status: Optional[Status] = None,
    ) -> None:
        self.path = Path(path)
        self.program = program or self.__annotations__["program"](
            post_init=post_init, status=status
        )

        if post_init:
            self.__post_init__()

    def __post_init__(self):
        ...

    def cmd(self, *args) -> subprocess.CompletedProcess:
        ...

    def new(self):
        ...

    def add(self, packages: list):
        ...

    def install(self):
        ...



from pathlib import Path

import typer
from rich.pretty import pprint

from pyvarium.config import Scope, settings

app = typer.Typer(no_args_is_help=True, help="Modify user settings for pyvarium.")


@app.command(name="list")
def _list() -> None:
    """List configuration."""
    pprint(settings.dict(), expand_all=True, indent_guides=False)


@app.command(name="set")
def _set(key: str, value: str, scope: Scope = Scope.local) -> None:
    """Set a key value pair for configuration, e.g.
    `pyvarium set poetry_exec /opt/poetry/bin/poetry`."""
    settings.__setattr__(key, value)
    settings.write(scope)
    _list()


@app.command()
def unset(key: str, scope: Scope = Scope.local) -> None:
    """Remove a custom settings from configuration (this reverts to the default, to
    disable a default set it to an empty string, e.g. `pyvarium set poetry_exec ""`)."""
    settings.__delattr__(key)
    settings.write(scope)
    _list()


@app.command()
def info() -> None:
    """Show information about the current configuration."""

    files = settings.settings_scopes()

    pprint(
        files,
        indent_guides=False,
        expand_all=True,
    )



from pathlib import Path
from typing import Optional

from pyvarium.installers.base import Environment, Program

PIPFILE = """[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]

[dev-packages]

"""


class Pipenv(Program):
    def __post_init__(self):
        self.env["PIPENV_VENV_IN_PROJECT"] = "1"


class PipenvEnvironment(Environment):
    program: Pipenv

    def __post_init__(self):
        self.program.cwd = self.path
        self.program.env["VIRTUAL_ENV"] = str((self.path / ".venv").absolute())
        self.program.env["PIPENV_PYTHON"] = str(
            (self.path / ".venv" / "bin" / "python").absolute()
        )

    def new(
        self,
        *,
        python_path: Optional[Path] = None,
    ):
        commands = []

        if python_path is not None:
            commands.extend([f"--python={str(python_path.absolute())}"])
        else:
            commands.extend(["--three"])

        self.path.mkdir(exist_ok=True, parents=True)

        (self.path / "Pipfile").write_text(PIPFILE)

        return self.program.cmd(*commands)

    def add(self, *packages):
        return self.program.cmd("install", *packages)

    def install(self):
        return self.program.cmd("install")

    def lock(self):
        return self.program.cmd("lock")



import json
import shutil
import subprocess
from pathlib import Path
from typing import List, Literal, Union, overload

import yaml
from loguru import logger

from pyvarium.installers.base import Environment, Program
from pyvarium.util import python_venv


def recursive_dict_update(d, u):
    for k, v in u.items():
        if isinstance(v, dict):
            r = recursive_dict_update(d.get(k, {}) if d is dict else {}, v)
            d[k] = r
        else:
            d[k] = v
    return d


def cmd_json_to_dict(cmd: subprocess.CompletedProcess) -> dict:
    return json.loads(cmd.stdout.decode())


class Spack(Program):
    def __post_init__(self):
        spack_dir = self.executable.parent.parent
        hooks_dir = spack_dir / "lib" / "spack" / "spack" / "hooks"

        if not hooks_dir.exists():  # pragma: no cover
            raise FileNotFoundError(
                f"Spack hooks directory does not exist: {hooks_dir}"
            )

        shutil.copy(
            python_venv.__file__,
            hooks_dir / "pyvarium_venv_activate.py",
        )


class SpackEnvironment(Environment):
    program: Spack

    def cmd(self, *args):
        return self.program.cmd("--env-dir", str(self.path), *args)

    def new(self, *, view_path: Path = Path(".venv")):
        commands = ["env", "create", "-d", str(self.path)]

        if view_path:
            if not view_path.is_absolute():
                view_path = self.path / view_path
            # Views are disabled here, as we set them manually via `set_config` to set
            # the link type to `run`
            commands.extend(["--without-view"])

        res = self.program.cmd(*commands)

        self.set_config({"spack": {"concretizer": {"unify": True}}})
        self.set_config(
            {
                "spack": {
                    "view": {
                        "default": {"root": str(view_path.resolve()), "link": "run"}
                    },
                    "concretizer": {"unify": True},
                }
            }
        )

        return res

    def init_view(self):
        res = self.cmd("env", "view", "regenerate")
        python_venv.setup_scripts(self.path / ".venv")
        return res

    def add(self, *packages):
        return self.cmd("add", *packages)

    def install(self):
        if not (self.path / "spack.lock").exists():
            logger.warning("No spack.lock file found, nothing will be installed")

        return self.cmd("install", "--only-concrete", "--no-add")

    # def spec(self, spec: str) -> dict:
    #     res = self.cmd("spec", "-I", "--reuse", "--json", spec)
    #     return cmd_json_to_dict(res)

    def concretize(self):
        return self.cmd("concretize", "--reuse")

    def find(self) -> dict:
        res = self.cmd("find", "--json")
        return cmd_json_to_dict(res)

    # def find_missing(self) -> dict:
    #     res = self.cmd(
    #         "find", "--show-concretized", "--deps", "--only-missing", "--json"
    #     )
    #     return cmd_json_to_dict(res)

    @overload
    def find_python_packages(self) -> List[dict]:
        ...

    @overload
    def find_python_packages(self, only_names: Literal[True]) -> List[str]:
        ...

    @overload
    def find_python_packages(self, only_names: Literal[False]) -> List[dict]:
        ...

    def find_python_packages(
        self, only_names: bool = False
    ) -> Union[List[str], List[dict]]:
        cmd = "PYTHONNOUSERSITE=True .venv/bin/python -m pip list --format json --disable-pip-version-check"
        res = subprocess.run(cmd, shell=True, capture_output=True, cwd=self.path)
        logger.debug(res)
        packages_json = res.stdout.decode().strip()

        if not packages_json:
            return []  # type: ignore

        packages_dict: List[dict] = json.loads(packages_json)

        if only_names:
            return [
                f"{p['name']}=={p['version']}"
                for p in packages_dict
                if p["name"] != "pip"
            ]
        else:
            return packages_dict

    def get_config(self) -> dict:
        return yaml.safe_load((self.path / "spack.yaml").read_text())

    def set_config(self, config: dict) -> None:
        current_config = self.get_config()
        new_config = recursive_dict_update(current_config, config)

        yaml.dump(new_config, (self.path / "spack.yaml").open("w"))



from pathlib import Path
from typing import Optional

import typer
from jinja2 import Template

from pyvarium.installers import spack

app = typer.Typer(help="Generate modulefile to load the environment.")

modulefile_template = Template(
    """#%Module 1.0
#
#  {{name}}
#

module-whatis  '{{name}} modulefile'
if { [ module-info mode load ] } {
{%- for path in paths %}
    prepend-path {{ path -}}
{% endfor %}
} elseif { [module-info mode remove] && ![module-info mode switch3] } {
{%- for path in paths %}
    remove-path {{path -}}
{% endfor %}
}
"""
)


@app.callback(invoke_without_command=True)
def main(
    path: Path = typer.Option(".", file_okay=False),
    name: Optional[str] = typer.Option("/".join(Path.cwd().parts[-2:])),
):
    path = Path(path).absolute()

    se = spack.SpackEnvironment(path)
    res = se.program.cmd("env", "activate", "--sh", str(path))

    commands = res.stdout.decode().split("\n")

    env_vars = {
        l.strip("export ").split("=")[0]: l.strip("export ").split("=")[1]
        for l in commands
        if l.startswith("export ")
    }

    paths = env_vars["PATH"].strip(";").split(":")
    if se.program.env["PATH"] in paths:
        paths.remove(se.program.env["PATH"])

    env_vars["PATH"] = ":".join(paths)
    env_vars["VIRTUAL_ENV"] = name

    paths = {
        k: ":".join(list(v.split(":"))).strip(";").strip(":")
        for k, v in env_vars.items()
    }

    modulefile = modulefile_template.render(
        name=name, paths=[f"{k} {v}" for k, v in paths.items()]
    )

    typer.echo(modulefile, color=False)



import venv
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Context:
    """Subsection of the context provided by venv.EnvBuilder.ensure_directories"""

    bin_path: str
    bin_name: str
    env_dir: str
    env_name: str
    prompt: str
    env_exe: str


def setup_scripts(spack_view_path: Path):
    spack_view_path = spack_view_path.absolute()
    eb = venv.EnvBuilder()
    context = Context(
        env_dir=str(spack_view_path),
        env_name=str(spack_view_path.parent.parent),
        prompt=f"({spack_view_path.parent.parent.name}) ",
        bin_path=str(spack_view_path / "bin"),
        bin_name="bin",
        env_exe=str(spack_view_path / "bin" / "python"),
    )

    eb.setup_scripts(context)  # type: ignore


def post_env_write(env):  # pragma: no cover
    # The scripts are written every time an env event occurs, as they almost always
    # overwrite/delete the existing scripts
    if "default" in env.views:
        env_path = Path(env.views["default"].root)
        if env_path.exists() and not (env_path / "bin" / "activate").exists():
            setup_scripts(env_path)



from pathlib import Path
from typing import List, Optional

import typer
from loguru import logger
from rich.status import Status

from pyvarium.installers import pipenv, spack

app = typer.Typer(no_args_is_help=True, help="Add packages via spack or pipenv.")


def main(
    path: Path,
    *,
    pipenv_add: Optional[List[str]] = None,
    spack_add: Optional[List[str]] = None,
):
    path = path.resolve()

    with Status("Spack add") as status:
        se = spack.SpackEnvironment(path, status=status)

        if spack_add:
            se.add(*spack_add)
            se.concretize()
            se.install()

    with Status("Pipenv add") as status:
        pe = pipenv.PipenvEnvironment(path, status=status)
        if se_python := se.find_python_packages(only_names=True):
            logger.info(f"Python packages in spack environment: {se_python}")
            pe.add(*se_python)

        if pipenv_add:
            pe.add(*pipenv_add)


@app.command(name="spack")
def _spack(
    packages: List[str],
    path: Path = typer.Option(".", file_okay=False),
):
    main(path, spack_add=packages)


@app.command(name="pipenv")
def _pipenv(
    packages: List[str],
    path: Path = typer.Option("", file_okay=False),
):
    main(path, pipenv_add=packages)



from pathlib import Path

import typer
from loguru import logger
from rich.status import Status

from pyvarium.installers import pipenv, spack

app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(path: Path = typer.Argument(..., file_okay=False)):
    """Create a new combined Spack and Pipenv environment."""

    if path.exists():
        logger.error(
            f"Directory already exists at path {path.absolute()}\n"
            "Remove it or use another path to continue."
        )
        raise typer.Exit(code=1)

    with Status("Spack setup") as status:
        se = spack.SpackEnvironment(path, status=status)
        se.new()
        se.add("python", "py-pip")
        se.concretize()
        se.install()

    with Status("Pipenv setup") as status:
        pe = pipenv.PipenvEnvironment(path, status=status)
        pe.new(python_path=se.path / ".venv" / "bin" / "python")
        if se_python := se.find_python_packages(only_names=True):
            pe.add(*se_python)
        pe.lock()



import shutil
import subprocess
import sys
from enum import Enum
from pathlib import Path

import typer
from loguru import logger
from rich.console import Console
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.prompt import Confirm

from . import add, config, install, modulegen, new, sync

app = typer.Typer()

CONSOLE = Console()


class LogLevel(str, Enum):
    """Enum of valid logging levels for console output."""

    trace = "TRACE"
    debug = "DEBUG"
    info = "INFO"
    warning = "WARNING"
    error = "ERROR"
    critical = "CRITICAL"

    @classmethod
    def _missing_(cls, name):
        for member in cls:
            if member.name.lower() == name:
                return member

        raise ValueError(f"{name} is not a valid log level")


def pre_checks():
    real_python_path = Path(sys.executable).resolve()
    external_dependencies = {
        "pipx": (shutil.which("pipx"), f"{real_python_path} -m pip install pipx"),
        "pipenv": (shutil.which("pipenv"), "pipx install pipenv"),
    }

    for name, (path, install_cmd) in external_dependencies.items():
        if path is None:
            CONSOLE.print(
                Markdown(
                    f"`{name}` is **required** for pyvarium to work but cannot be"
                    "installed as a direct dependency. Would you like to install it? "
                    f"This will run the command `{install_cmd}`."
                )
            )
            if Confirm.ask(f"Run `{install_cmd}`?"):
                subprocess.run(install_cmd, shell=True)
            else:
                raise FileNotFoundError(name)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    log_level: LogLevel = typer.Option(
        LogLevel.info, help="Pick which level of output to show", case_sensitive=False
    ),
):
    """Deploy mixed computational environments with dependencies and packages
    provided by Spack and Pipenv"""

    pre_checks()

    logger.remove()

    logger.configure(
        handlers=[
            {
                "sink": RichHandler(markup=True),
                "format": "{message}",
                "level": log_level,
            }
        ]
    )


app.add_typer(add.app, name="add")
app.add_typer(config.app, name="config")
app.add_typer(install.app, name="install")
app.add_typer(modulegen.app, name="modulegen")
app.add_typer(new.app, name="new")
app.add_typer(sync.app, name="sync")


if __name__ == "__main__":
    app()



from pathlib import Path

import typer
from rich.status import Status

from pyvarium.installers import pipenv, spack

app = typer.Typer(help="Sync Spack-managed packages with Pipenv.")


@app.callback(invoke_without_command=True)
def main(path: Path = typer.Option(".", file_okay=False)):
    path = path.resolve()

    with Status("Syncing Spack and Pipenv packages") as status:
        se = spack.SpackEnvironment(path, status=status)
        pe = pipenv.PipenvEnvironment(path, status=status)
        if se_python := se.find_python_packages(only_names=True):
            pe.add(*se_python)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyvarium', 'pyvarium.cli', 'pyvarium.installers', 'pyvarium.util']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'loguru>=0.5.3,<0.6.0',
 'pyaml>=21.8.3,<22.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'rich>=10.7.0,<11.0.0',
 'rtoml>=0.8.0,<0.9.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['pyvarium = pyvarium.cli:app']}

setup_kwargs = {
    'name': 'pyvarium',
    'version': '0.1.0',
    'description': 'Tool for managing mixed Spack and pip packages',
    'long_description': 'None',
    'author': 'Robert Rosca',
    'author_email': '32569096+RobertRosca@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0.0',
}


setup(**setup_kwargs)
